# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language import (
    PyLegendBoolean,
    PyLegendString,
    PyLegendInteger,
    PyLegendFloat,
    PyLegendNumber,
    PyLegendStrictDate,
    PyLegendDateTime,
    PyLegendDate,
)
from pylegend.core.language.legendql_api.legendql_api_custom_expressions import (
    LegendQLApiBoolean,
    LegendQLApiString,
    LegendQLApiInteger,
    LegendQLApiFloat,
    LegendQLApiNumber,
    LegendQLApiStrictDate,
    LegendQLApiDateTime,
    LegendQLApiDate,
    LegendQLApiPrimitive,
    LegendQLApiPartialFrame,
    LegendQLApiWindowReference,
)
from pylegend.core.language.shared.tds_row import AbstractTdsRow
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    Expression,
    FunctionCall,
    QualifiedName,
    IntegerLiteral
)
from pylegend.core.tds.tds_frame import PyLegendTdsFrame, FrameToPureConfig, FrameToSqlConfig

__all__: PyLegendSequence[str] = [
    "LegendQLApiTdsRow",
    "LegendQLApiLeadRow",
    "LegendQLApiLagRow",
    "LegendQLApiFirstRow",
    "LegendQLApiLastRow",
    "LegendQLApiNthRow",
]


class LegendQLApiTdsRow(AbstractTdsRow):

    def __init__(self, frame_name: str, frame: PyLegendTdsFrame) -> None:
        super().__init__(frame_name, frame)

    @staticmethod
    def from_tds_frame(frame_name: str, frame: PyLegendTdsFrame) -> "LegendQLApiTdsRow":
        return LegendQLApiTdsRow(frame_name=frame_name, frame=frame)

    def __getattr__(self, key: str) -> LegendQLApiPrimitive:
        return self[key]

    def get_boolean(self, column: str) -> LegendQLApiBoolean:
        return LegendQLApiBoolean(super().get_boolean(column))

    def get_string(self, column: str) -> LegendQLApiString:
        return LegendQLApiString(super().get_string(column))

    def get_number(self, column: str) -> LegendQLApiNumber:
        return LegendQLApiNumber(super().get_number(column))

    def get_integer(self, column: str) -> LegendQLApiInteger:
        return LegendQLApiInteger(super().get_integer(column))

    def get_float(self, column: str) -> LegendQLApiFloat:
        return LegendQLApiFloat(super().get_float(column))

    def get_date(self, column: str) -> LegendQLApiDate:
        return LegendQLApiDate(super().get_date(column))

    def get_datetime(self, column: str) -> LegendQLApiDateTime:
        return LegendQLApiDateTime(super().get_datetime(column))

    def get_strictdate(self, column: str) -> LegendQLApiStrictDate:
        return LegendQLApiStrictDate(super().get_strictdate(column))

    def __getitem__(self, item: str) -> LegendQLApiPrimitive:
        res = super().__getitem__(item)
        if isinstance(res, PyLegendBoolean):
            return LegendQLApiBoolean(res)
        if isinstance(res, PyLegendString):
            return LegendQLApiString(res)
        if isinstance(res, PyLegendInteger):
            return LegendQLApiInteger(res)
        if isinstance(res, PyLegendFloat):
            return LegendQLApiFloat(res)
        if isinstance(res, PyLegendNumber):
            return LegendQLApiNumber(res)
        if isinstance(res, PyLegendStrictDate):
            return LegendQLApiStrictDate(res)
        if isinstance(res, PyLegendDateTime):
            return LegendQLApiDateTime(res)
        if isinstance(res, PyLegendDate):
            return LegendQLApiDate(res)

        raise RuntimeError(f"Unhandled primitive type {type(res)} in LegendQL Api")


class LegendQLApiLeadRow(LegendQLApiTdsRow):
    __partial_frame: LegendQLApiPartialFrame
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            row: "LegendQLApiTdsRow"
    ) -> None:
        super().__init__(frame_name=row.get_frame_name(), frame=partial_frame.get_base_frame())
        self.__partial_frame = partial_frame
        self.__row = row

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"{self.__partial_frame.to_pure_expression(config)}->lead({self.__row.to_pure_expression(config)})"

    def column_sql_expression(
            self,
            column: str,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["lead"]),
            distinct=False,
            arguments=[super().column_sql_expression(column, frame_name_to_base_query_map, config)],
            filter_=None,
            window=None
        )


class LegendQLApiLagRow(LegendQLApiTdsRow):
    __partial_frame: LegendQLApiPartialFrame
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            row: "LegendQLApiTdsRow"
    ) -> None:
        super().__init__(frame_name=row.get_frame_name(), frame=partial_frame.get_base_frame())
        self.__partial_frame = partial_frame
        self.__row = row

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"{self.__partial_frame.to_pure_expression(config)}->lag({self.__row.to_pure_expression(config)})"

    def column_sql_expression(
            self,
            column: str,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["lag"]),
            distinct=False,
            arguments=[super().column_sql_expression(column, frame_name_to_base_query_map, config)],
            filter_=None,
            window=None
        )


class LegendQLApiFirstRow(LegendQLApiTdsRow):
    __partial_frame: LegendQLApiPartialFrame
    __window_ref: LegendQLApiWindowReference
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            window_ref: LegendQLApiWindowReference,
            row: "LegendQLApiTdsRow"
    ) -> None:
        super().__init__(frame_name=row.get_frame_name(), frame=partial_frame.get_base_frame())
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->first("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")

    def column_sql_expression(
            self,
            column: str,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["first_value"]),
            distinct=False,
            arguments=[super().column_sql_expression(column, frame_name_to_base_query_map, config)],
            filter_=None,
            window=None
        )


class LegendQLApiLastRow(LegendQLApiTdsRow):
    __partial_frame: LegendQLApiPartialFrame
    __window_ref: LegendQLApiWindowReference
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            window_ref: LegendQLApiWindowReference,
            row: "LegendQLApiTdsRow"
    ) -> None:
        super().__init__(frame_name=row.get_frame_name(), frame=partial_frame.get_base_frame())
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->last("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")

    def column_sql_expression(
            self,
            column: str,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["last_value"]),
            distinct=False,
            arguments=[super().column_sql_expression(column, frame_name_to_base_query_map, config)],
            filter_=None,
            window=None
        )


class LegendQLApiNthRow(LegendQLApiTdsRow):
    __partial_frame: LegendQLApiPartialFrame
    __window_ref: LegendQLApiWindowReference
    __row: "LegendQLApiTdsRow"
    __offset: int

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            window_ref: LegendQLApiWindowReference,
            row: "LegendQLApiTdsRow",
            offset: int
    ) -> None:
        super().__init__(frame_name=row.get_frame_name(), frame=partial_frame.get_base_frame())
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row
        self.__offset = offset

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (
            f"{self.__partial_frame.to_pure_expression(config)}->nth("
            f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)}, {self.__offset})"
        )

    def column_sql_expression(
            self,
            column: str,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["nth_value"]),
            distinct=False,
            arguments=[
                super().column_sql_expression(column, frame_name_to_base_query_map, config),
                IntegerLiteral(self.__offset)
            ],
            filter_=None,
            window=None
        )



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pylegend._typing import (
    PyLegendOptional,
    PyLegendUnion,
    PyLegendList,
    PyLegendDict,
    TYPE_CHECKING,
)
from pylegend.core.language.pandas_api.pandas_api_aggregate_specification import PyLegendAggInput
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitiveOrPythonPrimitive
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame

if TYPE_CHECKING:
    from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame


class PandasApiGroupbyTdsFrame:
    __base_frame: PandasApiBaseTdsFrame
    __by: PyLegendUnion[str, PyLegendList[str]]
    __level: PyLegendOptional[PyLegendUnion[str, int, PyLegendList[str]]]
    __as_index: bool
    __sort: bool
    __group_keys: bool
    __observed: bool
    __dropna: bool

    __grouping_column_name_list: PyLegendList[str]
    __selected_columns: PyLegendOptional[PyLegendList[str]]

    @classmethod
    def name(cls) -> str:
        return "groupby"  # pragma: no cover

    def __init__(
        self,
        base_frame: PandasApiBaseTdsFrame,
        by: PyLegendUnion[str, PyLegendList[str]],
        level: PyLegendOptional[PyLegendUnion[str, int, PyLegendList[str]]] = None,
        as_index: bool = False,
        sort: bool = True,
        group_keys: bool = False,
        observed: bool = False,
        dropna: bool = False,
    ) -> None:
        self.__base_frame = base_frame
        self.__by = by
        self.__level = level
        self.__as_index = as_index
        self.__sort = sort
        self.__group_keys = group_keys
        self.__observed = observed
        self.__dropna = dropna

        self.__selected_columns = None

        self.__validate()

    def __validate(self) -> None:

        if self.__level is not None:
            raise NotImplementedError(
                "The 'level' parameter of the groupby function is not supported yet. "
                "Please specify groupby column names using the 'by' parameter."
            )

        if self.__as_index is not False:
            raise NotImplementedError(
                f"The 'as_index' parameter of the groupby function must be False, "
                f"but got: {self.__as_index} (type: {type(self.__as_index).__name__})"
            )

        if self.__group_keys is not False:
            raise NotImplementedError(
                f"The 'group_keys' parameter of the groupby function must be False, "
                f"but got: {self.__group_keys} (type: {type(self.__group_keys).__name__})"
            )

        if self.__observed is not False:
            raise NotImplementedError(
                f"The 'observed' parameter of the groupby function must be False, "
                f"but got: {self.__observed} (type: {type(self.__observed).__name__})"
            )

        if self.__dropna is not False:
            raise NotImplementedError(
                f"The 'dropna' parameter of the groupby function must be False, "
                f"but got: {self.__dropna} (type: {type(self.__dropna).__name__})"
            )

        input_cols: PyLegendList[str] = []
        if isinstance(self.__by, str):
            input_cols = [self.__by]
        elif isinstance(self.__by, list):
            input_cols = self.__by
        else:
            raise TypeError(
                f"The 'by' parameter in groupby function must be a string or a list of strings."
                f"but got: {self.__by} (type: {type(self.__by).__name__})"
            )

        if len(input_cols) == 0:
            raise ValueError("The 'by' parameter in groupby function must contain at least one column name.")

        available_columns = {c.get_name() for c in self.__base_frame.columns()}
        missing_cols = [col for col in input_cols if col not in available_columns]

        if len(missing_cols) > 0:
            raise KeyError(
                f"Column(s) {missing_cols} in groupby function's provided columns list "
                f"do not exist in the current frame. "
                f"Current frame columns: {sorted(available_columns)}"
            )

        self.__grouping_column_name_list = input_cols.copy()

    def __getitem__(self, item: PyLegendUnion[str, PyLegendList[str]]) -> "PandasApiGroupbyTdsFrame":
        columns_to_select: PyLegendList[str] = []

        if isinstance(item, str):
            columns_to_select = [item]
        elif isinstance(item, list):
            columns_to_select = item
        else:
            raise TypeError(
                f"Column selection after groupby function must be a string or a list of strings, "
                f"but got: {item} (type: {type(item).__name__})"
            )

        if len(columns_to_select) == 0:
            raise ValueError("When performing column selection after groupby, at least one column must be selected.")

        available_columns = {c.get_name() for c in self.__base_frame.columns()}
        missing_cols = [col for col in columns_to_select if col not in available_columns]

        if len(missing_cols) > 0:
            raise KeyError(
                f"Column(s) {missing_cols} selected after groupby do not exist in the current frame. "
                f"Current frame columns: {sorted(available_columns)}"
            )

        new_frame = PandasApiGroupbyTdsFrame(
            base_frame=self.__base_frame,
            by=self.__by,
            level=self.__level,
            as_index=self.__as_index,
            sort=self.__sort,
            group_keys=self.__group_keys,
            observed=self.__observed,
            dropna=self.__dropna,
        )

        new_frame.__selected_columns = columns_to_select.copy()
        return new_frame

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def grouping_column_name_list(self) -> PyLegendList[str]:
        return self.__grouping_column_name_list.copy()

    def selected_columns(self) -> PyLegendOptional[PyLegendList[str]]:
        if self.__selected_columns is None:
            return None
        return self.__selected_columns.copy()

    def aggregate(
        self,
        func: PyLegendAggInput,
        axis: PyLegendUnion[int, str] = 0,
        *args: PyLegendPrimitiveOrPythonPrimitive,
        **kwargs: PyLegendPrimitiveOrPythonPrimitive,
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunctionTdsFrame
        from pylegend.core.tds.pandas_api.frames.functions.aggregate_function import AggregateFunction

        aggregated_result: PandasApiAppliedFunctionTdsFrame = PandasApiAppliedFunctionTdsFrame(
            AggregateFunction(self, func, axis, *args, **kwargs)
        )

        if self.__sort:
            from pylegend.core.tds.pandas_api.frames.functions.sort_values_function import SortValuesFunction

            aggregated_result = PandasApiAppliedFunctionTdsFrame(
                SortValuesFunction(
                    base_frame=aggregated_result,
                    by=self.grouping_column_name_list(),
                    axis=0,
                    ascending=True,
                    inplace=False,
                    kind=None,
                    na_position="last",
                    ignore_index=True,
                    key=None,
                )
            )

        return aggregated_result

    def agg(
        self,
        func: PyLegendAggInput,
        axis: PyLegendUnion[int, str] = 0,
        *args: PyLegendPrimitiveOrPythonPrimitive,
        **kwargs: PyLegendPrimitiveOrPythonPrimitive,
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunctionTdsFrame
        from pylegend.core.tds.pandas_api.frames.functions.aggregate_function import AggregateFunction

        return PandasApiAppliedFunctionTdsFrame(AggregateFunction(self, func, axis, *args, **kwargs))

    def sum(
        self,
        numeric_only: bool = False,
        min_count: int = 0,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
    ) -> "PandasApiTdsFrame":
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in sum function.")
        if min_count != 0:
            raise NotImplementedError(f"min_count must be 0 in sum function, but got: {min_count}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in sum function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in sum function.")
        return self.aggregate("sum", 0)

    def mean(
        self,
        numeric_only: bool = False,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
    ) -> "PandasApiTdsFrame":
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in mean function.")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in mean function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in mean function.")
        return self.aggregate("mean", 0)

    def min(
        self,
        numeric_only: bool = False,
        min_count: int = -1,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
    ) -> "PandasApiTdsFrame":
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in min function.")
        if min_count != -1:
            raise NotImplementedError(f"min_count must be -1 (default) in min function, but got: {min_count}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in min function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in min function.")
        return self.aggregate("min", 0)

    def max(
        self,
        numeric_only: bool = False,
        min_count: int = -1,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
    ) -> "PandasApiTdsFrame":
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in max function.")
        if min_count != -1:
            raise NotImplementedError(f"min_count must be -1 (default) in max function, but got: {min_count}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in max function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in max function.")
        return self.aggregate("max", 0)

    def std(
        self,
        ddof: int = 1,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
        numeric_only: bool = False,
    ) -> "PandasApiTdsFrame":
        if ddof != 1:
            raise NotImplementedError(f"Only ddof=1 (Sample Standard Deviation) is supported in std function, but got: {ddof}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in std function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in std function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in std function.")
        return self.aggregate("std", 0)

    def var(
        self,
        ddof: int = 1,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
        numeric_only: bool = False,
    ) -> "PandasApiTdsFrame":
        if ddof != 1:
            raise NotImplementedError(f"Only ddof=1 (Sample Variance) is supported in var function, but got: {ddof}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in var function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in var function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in var function.")
        return self.aggregate("var", 0)

    def count(self) -> "PandasApiTdsFrame":
        return self.aggregate("count", 0)



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import PyLegendList
__all__ = [
    "generate_pure_functional_call",
    "generate_pure_lambda",
    "escape_column_name",
    "expr_has_matching_start_and_end_parentheses",
]


def generate_pure_functional_call(
        func: str,
        params: PyLegendList[str],
        force_prefix: bool = False
) -> str:
    should_prefix = force_prefix or (len(params) == 0)

    updated_params: PyLegendList[str] = []
    for param in params:
        if param.startswith("(") and param.endswith(")"):
            if expr_has_matching_start_and_end_parentheses(param):
                updated_params.append(param[1:-1])
            else:
                updated_params.append(param)
        else:
            updated_params.append(param)

    if should_prefix:
        return f"{func}({', '.join(updated_params)})"
    else:
        return f"{params[0]}->{func}({', '.join(updated_params[1:])})"


def generate_pure_lambda(param_name: str, expr: str, wrap_in_braces: bool = True) -> str:
    lambda_code = param_name + " | " + (expr[1:-1] if expr_has_matching_start_and_end_parentheses(expr) else expr)
    return "{" + lambda_code + "}" if wrap_in_braces else lambda_code


def escape_column_name(name: str) -> str:
    return (name if name.isidentifier() else "'" + name.replace("'", "\\'") + "'")


def expr_has_matching_start_and_end_parentheses(expr: str) -> bool:
    if expr.startswith("(") and expr.endswith(")"):
        bracket_indices = [0]
        for (i, char) in enumerate(expr[1:-1]):
            if char == "(":
                bracket_indices.append(i + 1)
            elif char == ")":
                if bracket_indices:
                    bracket_indices.pop()
        return (len(bracket_indices) == 1) and (bracket_indices[0] == 0)
    else:
        return False



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABCMeta, abstractmethod
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig


__all__: PyLegendSequence[str] = [
    "PyLegendExpression",
    "PyLegendExpressionBooleanReturn",
    "PyLegendExpressionStringReturn",
    "PyLegendExpressionNumberReturn",
    "PyLegendExpressionIntegerReturn",
    "PyLegendExpressionFloatReturn",
    "PyLegendExpressionDateReturn",
    "PyLegendExpressionDateTimeReturn",
    "PyLegendExpressionStrictDateReturn",
]


class PyLegendExpression(metaclass=ABCMeta):
    @abstractmethod
    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        pass

    @abstractmethod
    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        pass

    def is_non_nullable(self) -> bool:
        return False


class PyLegendExpressionBooleanReturn(PyLegendExpression, metaclass=ABCMeta):
    pass


class PyLegendExpressionStringReturn(PyLegendExpression, metaclass=ABCMeta):
    pass


class PyLegendExpressionNumberReturn(PyLegendExpression, metaclass=ABCMeta):
    pass


class PyLegendExpressionIntegerReturn(PyLegendExpressionNumberReturn, metaclass=ABCMeta):
    pass


class PyLegendExpressionFloatReturn(PyLegendExpressionNumberReturn, metaclass=ABCMeta):
    pass


class PyLegendExpressionDateReturn(PyLegendExpression, metaclass=ABCMeta):
    pass


class PyLegendExpressionDateTimeReturn(PyLegendExpressionDateReturn, metaclass=ABCMeta):
    pass


class PyLegendExpressionStrictDateReturn(PyLegendExpressionDateReturn, metaclass=ABCMeta):
    pass



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from pylegend._typing import (
    PyLegendSequence,
    PyLegendList
)
from pylegend.core.tds.abstract.frames.input_tds_frame import InputTdsFrame, ExecutableInputTdsFrame, \
    NonExecutableInputTdsFrame
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.request.legend_client import LegendClient


__all__: PyLegendSequence[str] = [
    "PandasApiExecutableInputTdsFrame",
    "PandasApiNonExecutableInputTdsFrame",
    "PandasApiInputTdsFrame",
]


class PandasApiInputTdsFrame(PandasApiBaseTdsFrame, InputTdsFrame, metaclass=ABCMeta):

    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        super().__init__(columns=columns)

    def get_all_tds_frames(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return [self]


class PandasApiExecutableInputTdsFrame(PandasApiInputTdsFrame, ExecutableInputTdsFrame, metaclass=ABCMeta):
    __legend_client: LegendClient

    def __init__(self, legend_client: LegendClient, columns: PyLegendSequence[TdsColumn]) -> None:
        super().__init__(columns=columns)
        self.__legend_client = legend_client

    def get_legend_client(self) -> LegendClient:
        return self.__legend_client


class PandasApiNonExecutableInputTdsFrame(PandasApiInputTdsFrame, NonExecutableInputTdsFrame, metaclass=ABCMeta):

    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        super().__init__(columns=columns)



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
import pandas as pd
from pylegend._typing import (
    PyLegendSequence,
    PyLegendTypeVar,
    PyLegendOptional,
)
from pylegend.core.sql.metamodel import QuerySpecification
from pylegend.core.database.sql_to_string import (
    SqlToStringConfig,
    SqlToStringFormat
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig, PyLegendTdsFrame
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.result_handler import (
    ResultHandler,
    ToStringResultHandler,
)
from pylegend.extensions.tds.result_handler import (
    ToPandasDfResultHandler,
    PandasDfReadConfig,
)
__all__: PyLegendSequence[str] = [
    "BaseTdsFrame"
]

R = PyLegendTypeVar('R')


class BaseTdsFrame(PyLegendTdsFrame, metaclass=ABCMeta):
    __columns: PyLegendSequence[TdsColumn]

    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        col_names = [c.get_name() for c in columns]
        if len(col_names) != len(set(col_names)):
            cols = "[" + ", ".join([str(c) for c in columns]) + "]"
            raise ValueError(f"TdsFrame cannot have duplicated column names. Passed columns: {cols}")
        self.__columns = [c.copy() for c in columns]

    def columns(self) -> PyLegendSequence[TdsColumn]:
        return [c.copy() for c in self.__columns]

    @abstractmethod
    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        pass  # pragma: no cover

    @abstractmethod
    def get_all_tds_frames(self) -> PyLegendSequence["BaseTdsFrame"]:
        pass  # pragma: no cover

    def to_sql_query(self, config: FrameToSqlConfig = FrameToSqlConfig()) -> str:
        query = self.to_sql_query_object(config)
        sql_to_string_config = SqlToStringConfig(format_=SqlToStringFormat(pretty=config.pretty))
        return config.sql_to_string_generator().generate_sql_string(query, sql_to_string_config)

    @abstractmethod
    def to_pure(self, config: FrameToPureConfig) -> str:
        pass  # pragma: no cover

    def to_pure_query(self, config: FrameToPureConfig = FrameToPureConfig()) -> str:
        return self.to_pure(config)

    def execute_frame(
            self,
            result_handler: ResultHandler[R],
            chunk_size: PyLegendOptional[int] = None
    ) -> R:
        from pylegend.core.tds.abstract.frames.input_tds_frame import InputTdsFrame, ExecutableInputTdsFrame

        tds_frames = self.get_all_tds_frames()
        input_frames = [x for x in tds_frames if isinstance(x, InputTdsFrame)]

        non_exec_frames = [x for x in input_frames if not isinstance(x, ExecutableInputTdsFrame)]
        if non_exec_frames:
            raise ValueError(
                "Cannot execute frame as its built on top of non-executable input frames: [" +
                (", ".join([str(f) for f in non_exec_frames]) + "]")
            )

        exec_frames = [x for x in input_frames if isinstance(x, ExecutableInputTdsFrame)]

        all_legend_clients = []
        for e in exec_frames:
            c = e.get_legend_client()
            if c not in all_legend_clients:
                all_legend_clients.append(c)
        if len(all_legend_clients) > 1:
            raise ValueError(
                "Found tds frames with multiple legend_clients (which is not supported): [" +
                (", ".join([str(f) for f in all_legend_clients]) + "]")
            )
        legend_client = all_legend_clients[0]
        result = legend_client.execute_sql_string(self.to_sql_query(), chunk_size=chunk_size)
        return result_handler.handle_result(self, result)

    def execute_frame_to_string(
            self,
            chunk_size: PyLegendOptional[int] = None
    ) -> str:
        return self.execute_frame(ToStringResultHandler(), chunk_size)

    def execute_frame_to_pandas_df(
            self,
            chunk_size: PyLegendOptional[int] = None,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> pd.DataFrame:
        return self.execute_frame(
            ToPandasDfResultHandler(pandas_df_read_config),
            chunk_size
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from abc import ABCMeta, abstractmethod
import pandas as pd
from pylegend._typing import (
    PyLegendSequence,
    PyLegendTypeVar,
    PyLegendOptional,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.database.sql_to_string import SqlToStringGenerator
from pylegend.core.tds.result_handler import ResultHandler
from pylegend.extensions.tds.result_handler import PandasDfReadConfig

postgres_ext = 'pylegend.extensions.database.vendors.postgres.postgres_sql_to_string'
importlib.import_module(postgres_ext)

__all__: PyLegendSequence[str] = [
    "PyLegendTdsFrame",
    "FrameToSqlConfig",
    "FrameToPureConfig",
]


class FrameToSqlConfig:
    database_type: str
    pretty: bool

    __sql_to_string_generator: SqlToStringGenerator

    def __init__(
            self,
            database_type: str = "Postgres",
            pretty: bool = True
    ) -> None:
        self.database_type = database_type
        self.pretty = pretty

        self.__sql_to_string_generator = SqlToStringGenerator.find_sql_to_string_generator_for_db_type(
            self.database_type
        )

    def sql_to_string_generator(self) -> SqlToStringGenerator:
        return self.__sql_to_string_generator


class FrameToPureConfig:
    __pretty: bool
    __indent: str
    __indent_level: int

    def __init__(
            self,
            pretty: bool = True,
            ident: str = "  ",
            indent_level: int = 0,
    ) -> None:
        self.__pretty = pretty
        self.__indent = ident
        self.__indent_level = indent_level

    def push_indent(self, extra_indent_level: int = 1) -> "FrameToPureConfig":
        return FrameToPureConfig(self.__pretty, self.__indent, self.__indent_level + extra_indent_level)

    def separator(self, extra_indent_level: int = 0, return_space_if_not_pretty: bool = False) -> str:
        if self.__pretty:
            return "\n" + (self.__indent * (self.__indent_level + extra_indent_level))
        else:
            return " " if return_space_if_not_pretty else ""


R = PyLegendTypeVar('R')


class PyLegendTdsFrame(metaclass=ABCMeta):

    @abstractmethod
    def columns(self) -> PyLegendSequence[TdsColumn]:
        pass  # pragma: no cover

    @abstractmethod
    def to_sql_query(self, config: FrameToSqlConfig = FrameToSqlConfig()) -> str:
        pass  # pragma: no cover

    @abstractmethod
    def to_pure_query(self, config: FrameToPureConfig = FrameToPureConfig()) -> str:
        pass  # pragma: no cover

    @abstractmethod
    def execute_frame(
            self,
            result_handler: ResultHandler[R],
            chunk_size: PyLegendOptional[int] = None
    ) -> R:
        pass  # pragma: no cover

    @abstractmethod
    def execute_frame_to_string(
            self,
            chunk_size: PyLegendOptional[int] = None
    ) -> str:
        pass  # pragma: no cover

    @abstractmethod
    def execute_frame_to_pandas_df(
            self,
            chunk_size: PyLegendOptional[int] = None,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> pd.DataFrame:
        pass  # pragma: no cover



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from abc import ABCMeta
from pylegend._typing import (
    PyLegendSequence,
    PyLegendList,
)
from pylegend.core.sql.metamodel import (
    NamedArgumentExpression,
    StringLiteral,
)

__all__: PyLegendSequence[str] = [
    "ProjectCoordinates",
    "VersionedProjectCoordinates",
    "PersonalWorkspaceProjectCoordinates",
    "GroupWorkspaceProjectCoordinates",
]


class ProjectCoordinates(metaclass=ABCMeta):

    @abc.abstractmethod
    def sql_params(self) -> PyLegendList[NamedArgumentExpression]:
        pass


class VersionedProjectCoordinates(ProjectCoordinates):
    __group_id: str
    __artifact_id: str
    __version: str

    def __init__(self, group_id: str, artifact_id: str, version: str) -> None:
        self.__group_id = group_id
        self.__artifact_id = artifact_id
        self.__version = version

    def get_group_id(self) -> str:
        return self.__group_id

    def get_artifact_id(self) -> str:
        return self.__artifact_id

    def get_version(self) -> str:
        return self.__version

    def sql_params(self) -> PyLegendList[NamedArgumentExpression]:
        return [
            NamedArgumentExpression(
                name="coordinates",
                expression=StringLiteral(
                    value=f"{self.__group_id}:{self.__artifact_id}:{self.__version}",
                    quoted=False
                )
            )
        ]


class WorkspaceProjectCoordinates(ProjectCoordinates, metaclass=ABCMeta):
    __project_id: str

    def __init__(self, project_id: str) -> None:
        self.__project_id = project_id

    def get_project_id(self) -> str:
        return self.__project_id


class PersonalWorkspaceProjectCoordinates(WorkspaceProjectCoordinates):
    __workspace: str

    def __init__(self, project_id: str, workspace: str) -> None:
        super().__init__(project_id)
        self.__workspace = workspace

    def get_workspace(self) -> str:
        return self.__workspace

    def sql_params(self) -> PyLegendList[NamedArgumentExpression]:
        return [
            NamedArgumentExpression(
                name="project",
                expression=StringLiteral(value=self.get_project_id(), quoted=False)
            ),
            NamedArgumentExpression(
                name="workspace",
                expression=StringLiteral(value=self.__workspace, quoted=False)
            )
        ]


class GroupWorkspaceProjectCoordinates(WorkspaceProjectCoordinates):
    __group_workspace: str

    def __init__(self, project_id: str, group_workspace: str) -> None:
        super().__init__(project_id)
        self.__group_workspace = group_workspace

    def get_group_workspace(self) -> str:
        return self.__group_workspace

    def sql_params(self) -> PyLegendList[NamedArgumentExpression]:
        return [
            NamedArgumentExpression(
                name="project",
                expression=StringLiteral(value=self.get_project_id(), quoted=False)
            ),
            NamedArgumentExpression(
                name="groupWorkspace",
                expression=StringLiteral(value=self.__group_workspace, quoted=False)
            )
        ]



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language.shared.expression import (
    PyLegendExpression,
    PyLegendExpressionBooleanReturn,
    PyLegendExpressionStringReturn
)
from pylegend.core.language.shared.helpers import generate_pure_functional_call
from pylegend.core.language.shared.operations.binary_expression import PyLegendBinaryExpression
from pylegend.core.language.shared.operations.unary_expression import PyLegendUnaryExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
    ComparisonExpression,
    ComparisonOperator,
    IsNullPredicate,
    IsNotNullPredicate,
    Cast,
    ColumnType
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig


__all__: PyLegendSequence[str] = [
    "PyLegendPrimitiveEqualsExpression",
    "PyLegendPrimitiveNotEqualsExpression",
    "PyLegendIsEmptyExpression",
    "PyLegendIsNotEmptyExpression",
    "PyLegendPrimitiveToStringExpression"
]


class PyLegendPrimitiveEqualsExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.EQUAL)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} == {op2_expr})"

    def __init__(self, operand1: PyLegendExpression, operand2: PyLegendExpression) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendPrimitiveEqualsExpression.__to_sql_func,
            PyLegendPrimitiveEqualsExpression.__to_pure_func,
            non_nullable=True,
        )


class PyLegendPrimitiveNotEqualsExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.NOT_EQUAL)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} != {op2_expr})"

    def __init__(self, operand1: PyLegendExpression, operand2: PyLegendExpression) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendPrimitiveNotEqualsExpression.__to_sql_func,
            PyLegendPrimitiveNotEqualsExpression.__to_pure_func,
            non_nullable=True,
        )


class PyLegendIsEmptyExpression(PyLegendUnaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return IsNullPredicate(expression)

    @staticmethod
    def __to_pure_func(op_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("isEmpty", [op_expr])

    def __init__(self, operand: PyLegendExpression) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendIsEmptyExpression.__to_sql_func,
            PyLegendIsEmptyExpression.__to_pure_func,
            non_nullable=True,
        )


class PyLegendIsNotEmptyExpression(PyLegendUnaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return IsNotNullPredicate(expression)

    @staticmethod
    def __to_pure_func(op_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("isNotEmpty", [op_expr])

    def __init__(self, operand: PyLegendExpression) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendIsNotEmptyExpression.__to_sql_func,
            PyLegendIsNotEmptyExpression.__to_pure_func,
            non_nullable=True,
        )


class PyLegendPrimitiveToStringExpression(PyLegendUnaryExpression, PyLegendExpressionStringReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return Cast(expression, ColumnType(name="TEXT", parameters=[]))

    @staticmethod
    def __to_pure_func(op_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("toString", [op_expr])

    def __init__(self, operand: PyLegendExpression) -> None:
        PyLegendExpressionStringReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendPrimitiveToStringExpression.__to_sql_func,
            PyLegendPrimitiveToStringExpression.__to_pure_func,
            non_nullable=True,
            operand_needs_to_be_non_nullable=True,
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
    PyLegendOptional
)
from pylegend.core.language.shared.literal_expressions import PyLegendIntegerLiteralExpression
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitive
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.float import PyLegendFloat
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
from pylegend.core.language.shared.expression import PyLegendExpressionStringReturn
from pylegend.core.language.shared.literal_expressions import PyLegendStringLiteralExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.language.shared.operations.string_operation_expressions import (
    PyLegendStringLengthExpression,
    PyLegendStringStartsWithExpression,
    PyLegendStringEndsWithExpression,
    PyLegendStringContainsExpression,
    PyLegendStringUpperExpression,
    PyLegendStringLowerExpression,
    PyLegendStringLTrimExpression,
    PyLegendStringRTrimExpression,
    PyLegendStringBTrimExpression,
    PyLegendStringPosExpression,
    PyLegendStringParseIntExpression,
    PyLegendStringParseFloatExpression,
    PyLegendStringConcatExpression,
    PyLegendStringLessThanExpression,
    PyLegendStringLessThanEqualExpression,
    PyLegendStringGreaterThanExpression,
    PyLegendStringGreaterThanEqualExpression,
    PyLegendStringParseBooleanExpression,
    PyLegendStringParseDateTimeExpression,
    PyLegendStringAsciiExpression,
    PyLegendStringDecodeBase64Expression,
    PyLegendStringEncodeBase64Expression,
    PyLegendStringReverseExpression,
    PyLegendStringToLowerFirstCharacterExpression,
    PyLegendStringToUpperFirstCharacterExpression,
    PyLegendStringLeftExpression,
    PyLegendStringRightExpression,
    PyLegendStringSubStringExpression,
    PyLegendStringReplaceExpression,
    PyLegendStringLpadExpression,
    PyLegendStringRpadExpression,
    PyLegendStringSplitPartExpression,
    PyLegendStringFullMatchExpression,
    PyLegendStringRepeatStringExpression,
    PyLegendStringMatchExpression
)

__all__: PyLegendSequence[str] = [
    "PyLegendString"
]


class PyLegendString(PyLegendPrimitive):
    __value: PyLegendExpressionStringReturn

    def __init__(
            self,
            value: PyLegendExpressionStringReturn
    ) -> None:
        self.__value = value

    def len(self) -> PyLegendInteger:
        return PyLegendInteger(PyLegendStringLengthExpression(self.__value))

    def length(self) -> PyLegendInteger:
        return self.len()

    def startswith(self, prefix: str) -> PyLegendBoolean:
        PyLegendString.__validate_param_to_be_str(prefix, "startswith prefix parameter")
        return PyLegendBoolean(
            PyLegendStringStartsWithExpression(self.__value, PyLegendStringLiteralExpression(prefix))
        )

    def endswith(self, suffix: str) -> PyLegendBoolean:
        PyLegendString.__validate_param_to_be_str(suffix, "endswith suffix parameter")
        return PyLegendBoolean(
            PyLegendStringEndsWithExpression(self.__value, PyLegendStringLiteralExpression(suffix))
        )

    def contains(self, other: str) -> PyLegendBoolean:
        PyLegendString.__validate_param_to_be_str(other, "contains/in other parameter")
        return PyLegendBoolean(
            PyLegendStringContainsExpression(self.__value, PyLegendStringLiteralExpression(other))
        )

    def upper(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringUpperExpression(self.__value))

    def lower(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringLowerExpression(self.__value))

    def lstrip(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringLTrimExpression(self.__value))

    def rstrip(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringRTrimExpression(self.__value))

    def strip(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringBTrimExpression(self.__value))

    def index_of(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendInteger":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "Index_of parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendInteger(PyLegendStringPosExpression(self.__value, other_op))

    def index(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendInteger":
        return self.index_of(other)

    def parse_int(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendStringParseIntExpression(self.__value))

    def parse_integer(self) -> "PyLegendInteger":
        return self.parse_int()

    def parse_float(self) -> "PyLegendFloat":
        return PyLegendFloat(PyLegendStringParseFloatExpression(self.__value))

    def parse_boolean(self) -> "PyLegendBoolean":
        return PyLegendBoolean(PyLegendStringParseBooleanExpression(self.__value))

    def parse_datetime(self) -> "PyLegendDateTime":
        return PyLegendDateTime(PyLegendStringParseDateTimeExpression(self.__value))

    def ascii(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendStringAsciiExpression(self.__value))

    def b64decode(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringDecodeBase64Expression(self.__value))

    def b64encode(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringEncodeBase64Expression(self.__value))

    def reverse(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringReverseExpression(self.__value))

    def to_lower_first_character(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringToLowerFirstCharacterExpression(self.__value))

    def to_upper_first_character(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringToUpperFirstCharacterExpression(self.__value))

    def left(self, count: PyLegendUnion[int, "PyLegendInteger"]) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_int_or_int_expr(count, "left parameter")
        count_op = PyLegendIntegerLiteralExpression(count) if isinstance(count, int) else count.value()
        return PyLegendString(PyLegendStringLeftExpression(self.__value, count_op))

    def right(self, count: PyLegendUnion[int, "PyLegendInteger"]) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_int_or_int_expr(count, "right parameter")
        count_op = PyLegendIntegerLiteralExpression(count) if isinstance(count, int) else count.value()
        return PyLegendString(PyLegendStringRightExpression(self.__value, count_op))

    def substring(
            self,
            start: PyLegendUnion[int, "PyLegendInteger"],
            end: PyLegendOptional[PyLegendUnion[int, "PyLegendInteger"]] = None) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_int_or_int_expr(start, "substring start parameter")
        start_op = PyLegendIntegerLiteralExpression(start) if isinstance(start, int) else start.value()
        if end is None:
            return PyLegendString(PyLegendStringSubStringExpression([self.__value, start_op]))
        PyLegendString.__validate_param_to_be_int_or_int_expr(end, "substring end parameter")
        end_op = PyLegendIntegerLiteralExpression(end) if isinstance(end, int) else end.value()
        return PyLegendString(PyLegendStringSubStringExpression([self.__value, start_op, end_op]))

    def replace(
            self,
            to_replace: PyLegendUnion[str, "PyLegendString"],
            replacement: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_str_or_str_expr(to_replace, "replace to_replace parameter")
        to_replace_op = PyLegendStringLiteralExpression(to_replace) if isinstance(to_replace,
                                                                                  str) else to_replace.__value
        PyLegendString.__validate_param_to_be_str_or_str_expr(replacement, "replace replacement parameter")
        replacement_op = PyLegendStringLiteralExpression(replacement) if isinstance(replacement,
                                                                                    str) else replacement.__value
        return PyLegendString(PyLegendStringReplaceExpression([self.__value, to_replace_op, replacement_op]))

    def rjust(
            self,
            length: PyLegendUnion[int, "PyLegendInteger"],
            fill_char: PyLegendUnion[str, "PyLegendString"] = ' ') -> "PyLegendString":
        PyLegendString.__validate_param_to_be_int_or_int_expr(length, "rjust length parameter")
        length_op = PyLegendIntegerLiteralExpression(length) if isinstance(length, int) else length.value()
        fill_char_op = PyLegendStringLiteralExpression(fill_char) if isinstance(fill_char, str) else fill_char.__value
        return PyLegendString(PyLegendStringLpadExpression([self.__value, length_op, fill_char_op]))

    def ljust(
            self,
            length: PyLegendUnion[int, "PyLegendInteger"],
            fill_char: PyLegendUnion[str, "PyLegendString"] = ' ') -> "PyLegendString":
        PyLegendString.__validate_param_to_be_int_or_int_expr(length, "ljust length parameter")
        length_op = PyLegendIntegerLiteralExpression(length) if isinstance(length, int) else length.value()
        fill_char_op = PyLegendStringLiteralExpression(fill_char) if isinstance(fill_char, str) else fill_char.__value
        return PyLegendString(PyLegendStringRpadExpression([self.__value, length_op, fill_char_op]))

    def split_part(
            self,
            sep: PyLegendUnion[str, "PyLegendString"],
            part: PyLegendUnion[int, "PyLegendInteger"]) -> "PyLegendString | None":
        PyLegendString.__validate_param_to_be_str_or_str_expr(sep, "split_part sep parameter")
        sep_op = PyLegendStringLiteralExpression(sep) if isinstance(sep, str) else sep.value()
        PyLegendString.__validate_param_to_be_int_or_int_expr(part, "split_part part parameter")
        part_op = PyLegendIntegerLiteralExpression(part) if isinstance(part, int) else part.value()
        return PyLegendString(PyLegendStringSplitPartExpression([self.__value, sep_op, part_op]))

    def full_match(self, pattern: PyLegendUnion[str, "PyLegendString"]) -> PyLegendBoolean:
        PyLegendString.__validate_param_to_be_str_or_str_expr(pattern, "full_match parameter")
        pattern_op = PyLegendStringLiteralExpression(pattern) if isinstance(pattern, str) else pattern.__value
        return PyLegendBoolean(PyLegendStringFullMatchExpression(self.__value, pattern_op))

    def match(self, pattern: PyLegendUnion[str, "PyLegendString"]) -> PyLegendBoolean:
        PyLegendString.__validate_param_to_be_str_or_str_expr(pattern, "match parameter")  # pragma: no cover
        pattern_op = PyLegendStringLiteralExpression(pattern) if isinstance(
            pattern, str) else pattern.__value  # pragma: no cover
        return PyLegendBoolean(PyLegendStringMatchExpression(self.__value, pattern_op))  # pragma: no cover

    def repeat_string(self, times: PyLegendUnion[int, "PyLegendInteger"]) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_int_or_int_expr(times, "repeatString parameter")
        times_op = PyLegendIntegerLiteralExpression(times) if isinstance(times, int) else times.value()
        return PyLegendString(PyLegendStringRepeatStringExpression(self.__value, times_op))

    def __add__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String plus (+) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendString(PyLegendStringConcatExpression(self.__value, other_op))

    def __radd__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String plus (+) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendString(PyLegendStringConcatExpression(other_op, self.__value))

    def __lt__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendBoolean":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String less than (<) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendBoolean(PyLegendStringLessThanExpression(self.__value, other_op))

    def __le__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendBoolean":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String less than equal (<=) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendBoolean(PyLegendStringLessThanEqualExpression(self.__value, other_op))

    def __gt__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendBoolean":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String greater than (>) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendBoolean(PyLegendStringGreaterThanExpression(self.__value, other_op))

    def __ge__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendBoolean":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String greater than equal (>=) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendBoolean(PyLegendStringGreaterThanEqualExpression(self.__value, other_op))

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return self.__value.to_sql_expression(frame_name_to_base_query_map, config)

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return self.__value.to_pure_expression(config)

    def value(self) -> PyLegendExpressionStringReturn:
        return self.__value

    @staticmethod
    def __validate_param_to_be_str_or_str_expr(param: PyLegendUnion[str, "PyLegendString"], desc: str) -> None:
        if not isinstance(param, (str, PyLegendString)):
            raise TypeError(desc + " should be a str or a string expression (PyLegendString)."
                                   " Got value " + str(param) + " of type: " + str(type(param)))

    @staticmethod
    def __validate_param_to_be_str(param: str, desc: str) -> None:
        if not isinstance(param, str):
            raise TypeError(desc + " should be a str."
                                   " Got value " + str(param) + " of type: " + str(type(param)))

    @staticmethod
    def __validate_param_to_be_int_or_int_expr(param: PyLegendUnion[int, "PyLegendInteger"], desc: str) -> None:
        if not isinstance(param, (int, PyLegendInteger)):
            raise TypeError(desc + " should be a int or a int expression (PyLegendInteger)."
                                   " Got value " + str(param) + " of type: " + str(type(param)))



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

from pylegend.core.language import PyLegendBoolean, PyLegendPrimitiveOrPythonPrimitive, PyLegendPrimitiveCollection, \
    PyLegendPrimitive
from pylegend.core.language.legendql_api.legendql_api_custom_expressions import (
    LegendQLApiPrimitive,
    LegendQLApiSortInfo,
    LegendQLApiWindow,
    LegendQLApiPartialFrame,
    LegendQLApiWindowReference,
    LegendQLApiWindowFrame,
)
from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow
from pylegend.core.tds.tds_frame import (
    PyLegendTdsFrame
)
from pylegend._typing import (
    PyLegendSequence,
    PyLegendCallable,
    PyLegendUnion,
    PyLegendList,
    PyLegendOptional,
    PyLegendTuple,
)

__all__: PyLegendSequence[str] = [
    "LegendQLApiTdsFrame"
]


class LegendQLApiTdsFrame(PyLegendTdsFrame, metaclass=ABCMeta):

    @abstractmethod
    def head(self, count: int = 5) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def limit(self, count: int = 5) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def drop(self, count: int = 5) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def slice(self, start_row: int, end_row_exclusive: int) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def distinct(
            self,
            columns: PyLegendOptional[PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                ]
            ]] = None
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def select(
            self,
            columns: PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def sort(
            self,
            sort_infos: PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[
                        LegendQLApiPrimitive,
                        LegendQLApiSortInfo,
                        PyLegendList[PyLegendUnion[LegendQLApiPrimitive, LegendQLApiSortInfo]],
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def concatenate(self, other: "LegendQLApiTdsFrame") -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def filter(
            self,
            filter_function: PyLegendCallable[[LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def rename(
            self,
            column_renames: PyLegendUnion[
                PyLegendTuple[str, str],
                PyLegendList[PyLegendTuple[str, str]],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[
                        PyLegendTuple[LegendQLApiPrimitive, str],
                        PyLegendList[PyLegendTuple[LegendQLApiPrimitive, str]]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def extend(
            self,
            extend_columns: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]
                ],
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                    PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                ],
                PyLegendList[
                    PyLegendUnion[
                        PyLegendTuple[
                            str,
                            PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]
                        ],
                        PyLegendTuple[
                            str,
                            PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                            PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                        ]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def join(
            self,
            other: "LegendQLApiTdsFrame",
            join_condition: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ],
            join_type: str = 'LEFT_OUTER'
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def inner_join(
            self,
            other: "LegendQLApiTdsFrame",
            join_condition: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def left_join(
            self,
            other: "LegendQLApiTdsFrame",
            join_condition: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def right_join(
            self,
            other: "LegendQLApiTdsFrame",
            join_condition: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def as_of_join(
            self,
            other: "LegendQLApiTdsFrame",
            match_function: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ],
            join_condition: PyLegendOptional[
                PyLegendCallable[[LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]]
            ] = None
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def group_by(
            self,
            grouping_columns: PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                ]
            ],
            aggregate_specifications: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                    PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                ],
                PyLegendList[
                    PyLegendTuple[
                        str,
                        PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                        PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def window(
            self,
            partition_by: PyLegendOptional[
                PyLegendUnion[
                    str,
                    PyLegendList[str],
                    PyLegendCallable[
                        [LegendQLApiTdsRow],
                        PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                    ]
                ]
            ] = None,
            order_by: PyLegendOptional[
                PyLegendUnion[
                    str,
                    PyLegendList[str],
                    PyLegendCallable[
                        [LegendQLApiTdsRow],
                        PyLegendUnion[
                            LegendQLApiPrimitive,
                            LegendQLApiSortInfo,
                            PyLegendList[PyLegendUnion[LegendQLApiPrimitive, LegendQLApiSortInfo]],
                        ]
                    ]
                ]
            ] = None,
            frame: PyLegendOptional[LegendQLApiWindowFrame] = None
    ) -> LegendQLApiWindow:
        pass

    @abstractmethod
    def window_extend(
            self,
            window: LegendQLApiWindow,
            extend_columns: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[
                        [LegendQLApiPartialFrame, LegendQLApiWindowReference, LegendQLApiTdsRow],
                        PyLegendPrimitiveOrPythonPrimitive
                    ]
                ],
                PyLegendTuple[
                    str,
                    PyLegendCallable[
                        [LegendQLApiPartialFrame, LegendQLApiWindowReference, LegendQLApiTdsRow],
                        PyLegendPrimitiveOrPythonPrimitive
                    ],
                    PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                ],
                PyLegendList[
                    PyLegendUnion[
                        PyLegendTuple[
                            str,
                            PyLegendCallable[
                                [LegendQLApiPartialFrame, LegendQLApiWindowReference, LegendQLApiTdsRow],
                                PyLegendPrimitiveOrPythonPrimitive
                            ]
                        ],
                        PyLegendTuple[
                            str,
                            PyLegendCallable[
                                [LegendQLApiPartialFrame, LegendQLApiWindowReference, LegendQLApiTdsRow],
                                PyLegendPrimitiveOrPythonPrimitive
                            ],
                            PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                        ]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass

    @abstractmethod
    def project(
            self,
            project_columns: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]
                ],
                PyLegendList[
                    PyLegendTuple[
                        str,
                        PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def rows(
            self,
            start: PyLegendUnion[str, int],
            end: PyLegendUnion[str, int]) -> LegendQLApiWindowFrame:
        pass  # pragma: no cover

    @abstractmethod
    def range(
            self,
            *,
            number_start: PyLegendOptional[PyLegendUnion[str, int, float]] = None,
            number_end: PyLegendOptional[PyLegendUnion[str, int, float]] = None,
            duration_start: PyLegendOptional[PyLegendUnion[str, int, float]] = None,
            duration_start_unit: PyLegendOptional[str] = None,
            duration_end: PyLegendOptional[PyLegendUnion[str, int, float]] = None,
            duration_end_unit: PyLegendOptional[str] = None) -> LegendQLApiWindowFrame:
        pass  # pragma: no cover



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendUnion,
    PyLegendOptional,
    PyLegendCallable
)
from pylegend.core.tds.legendql_api.frames.legendql_api_applied_function_tds_frame import LegendQLApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame
from pylegend.core.language.legendql_api.legendql_api_custom_expressions import LegendQLApiPrimitive
from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow
from pylegend.core.tds.legendql_api.frames.functions.legendql_api_function_helpers import infer_columns_from_frame
from pylegend.core.language.shared.helpers import escape_column_name

__all__: PyLegendSequence[str] = [
    "LegendQLApiDistinctFunction"
]


class LegendQLApiDistinctFunction(LegendQLApiAppliedFunction):
    __base_frame: LegendQLApiBaseTdsFrame
    __column_name_list: PyLegendOptional[PyLegendList[str]]

    @classmethod
    def name(cls) -> str:
        return "distinct"

    def __init__(
            self,
            base_frame: LegendQLApiBaseTdsFrame,
            columns: PyLegendOptional[PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                ]
            ]] = None
    ) -> None:
        self.__base_frame = base_frame
        self.__column_name_list = infer_columns_from_frame(base_frame, columns,
                                                           "'distinct' function 'columns'") if columns is not None else None

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (base_query.offset is not None) or (base_query.limit is not None) or (
                self.__column_name_list is not None)

        quoted_columns = None
        if self.__column_name_list is not None:
            db_extension = config.sql_to_string_generator().get_db_extension()
            quoted_columns = [
                db_extension.quote_identifier(col)
                for col in self.__column_name_list
            ]

        new_query = (
            create_sub_query(base_query, config, "root", quoted_columns) if should_create_sub_query else
            copy_query(base_query)
        )
        new_query.select.distinct = True

        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        columns_expr = (
            f"~[{', '.join(map(escape_column_name, self.__column_name_list))}]"
            if self.__column_name_list else ""
        )
        return (f"{self.__base_frame.to_pure(config)}{config.separator(1)}"
                f"->distinct({columns_expr})")

    def base_frame(self) -> LegendQLApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList[LegendQLApiBaseTdsFrame]:
        return []

    def calculate_columns(self) -> PyLegendSequence[TdsColumn]:
        if not self.__column_name_list:
            return [c.copy() for c in self.__base_frame.columns()]

        base_columns = self.__base_frame.columns()
        new_columns = []
        for name in self.__column_name_list:
            for col in base_columns:
                if col.get_name() == name:
                    new_columns.append(col.copy())
                    break

        return new_columns

    def validate(self) -> bool:
        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.tds.legendql_api.frames.legendql_api_input_tds_frame import (
    LegendQLApiNonExecutableInputTdsFrame,
)
from pylegend.extensions.tds.abstract.csv_tds_frame import CsvInputFrameAbstract

__all__: PyLegendSequence[str] = [
    "LegendQLApiCsvNonExecutableInputTdsFrame",
]


class LegendQLApiCsvNonExecutableInputTdsFrame(
    CsvInputFrameAbstract,
    LegendQLApiNonExecutableInputTdsFrame
):

    def __init__(
            self,
            csv_string: str) -> None:
        CsvInputFrameAbstract.__init__(self, csv_string=csv_string)
        LegendQLApiNonExecutableInputTdsFrame.__init__(self, columns=self.columns())



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce

from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    PyLegendList,
    PyLegendUnion,
)
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    LogicalBinaryType,
    LogicalBinaryExpression, BooleanLiteral,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
    PandasApiAppliedFunction,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import (
    PandasApiBaseTdsFrame,
)
from pylegend.core.tds.sql_query_helpers import copy_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig

__all__: PyLegendSequence[str] = ["PandasApiDropnaFunction"]


class PandasApiDropnaFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __axis: PyLegendUnion[int, str]
    __how: str
    __thresh: PyLegendOptional[int]
    __subset: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]]
    __inplace: bool
    __ignore_index: bool

    @classmethod
    def name(cls) -> str:
        return "dropna"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            axis: PyLegendUnion[int, str],
            how: str,
            thresh: PyLegendOptional[int],
            subset: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]],
            inplace: bool,
            ignore_index: bool
    ) -> None:
        self.__base_frame = base_frame
        self.__axis = axis
        self.__how = how
        self.__thresh = thresh
        self.__subset = subset
        self.__inplace = inplace
        self.__ignore_index = ignore_index

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        new_query = copy_query(base_query)

        if self.__subset is not None:
            cols_to_check = self.__subset
        else:
            cols_to_check = [c.get_name() for c in self.__base_frame.columns()]

        if not cols_to_check:
            if self.__how == 'all':
                new_query.where = BooleanLiteral(value=False)
            return new_query

        tds_row = PandasApiTdsRow.from_tds_frame("c", self.__base_frame)

        filter_expr = None

        conditions = [tds_row[col].is_not_null() for col in cols_to_check]
        if self.__how == "any":
            filter_expr = reduce(lambda x, y: x & y, conditions)
        else:  # "all"
            filter_expr = reduce(lambda x, y: x | y, conditions)

        sql_expr = filter_expr.to_sql_expression({"c": new_query}, config)
        if new_query.where is None:
            new_query.where = sql_expr
        else:
            new_query.where = LogicalBinaryExpression(
                type_=LogicalBinaryType.AND,
                left=new_query.where,
                right=sql_expr
            )

        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        base_pure = self.__base_frame.to_pure(config)
        if self.__subset is not None:
            cols_to_check = self.__subset
        else:
            cols_to_check = [c.get_name() for c in self.__base_frame.columns()]

        if not cols_to_check:
            if self.__how == 'all':
                return f"{base_pure}{config.separator(1)}->filter(c|1!=1)"
            return base_pure

        tds_row = PandasApiTdsRow.from_tds_frame("c", self.__base_frame)
        conditions = [tds_row[col].is_not_null() for col in cols_to_check]

        if self.__how == "any":
            filter_expr = reduce(lambda x, y: x & y, conditions)
        else:  # "all"
            filter_expr = reduce(lambda x, y: x | y, conditions)

        pure_expr = filter_expr.to_pure_expression(config)
        return f"{base_pure}{config.separator(1)}->filter(c|{pure_expr})"

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        if self.__axis not in (0, 1, "index", "columns"):
            raise ValueError(f"No axis named {self.__axis} for object type TdsFrame")
        if self.__axis in (1, "columns"):
            raise NotImplementedError("axis=1 is not supported yet in Pandas API dropna")

        if self.__thresh is not None:
            raise NotImplementedError("thresh parameter is not supported yet in Pandas API dropna")

        if self.__how not in ("any", "all"):
            raise ValueError(f"invalid how option: {self.__how}")

        if self.__subset is not None:
            if not isinstance(self.__subset, (list, tuple, set)):
                raise TypeError(f"subset must be a list, tuple or set of column names. Got {type(self.__subset)}")
            valid_cols = {c.get_name() for c in self.__base_frame.columns()}
            invalid_cols = [s for s in self.__subset if s not in valid_cols]
            if invalid_cols:
                raise KeyError(f"{invalid_cols}")

        if self.__inplace:
            raise NotImplementedError("inplace=True is not supported yet in Pandas API dropna")

        if self.__ignore_index:
            raise NotImplementedError("ignore_index=True is not supported yet in Pandas API dropna")

        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendList
)
from pylegend.core.language import PyLegendBoolean
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    LogicalBinaryExpression,
    LogicalBinaryType,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig

__all__: PyLegendSequence[str] = ["PandasApiFilteringFunction"]


class PandasApiFilteringFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __filter_expr: PyLegendBoolean

    @classmethod
    def name(cls) -> str:
        return "boolean_filter"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            filter_expr: PyLegendBoolean
    ) -> None:
        self.__base_frame = base_frame
        self.__filter_expr = filter_expr

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (len(base_query.groupBy) > 0) or \
                                  (base_query.offset is not None) or (base_query.limit is not None)
        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query else
            copy_query(base_query)
        )

        sql_expr = self.__filter_expr.to_sql_expression({"c": new_query}, config)

        if new_query.where is None:
            new_query.where = sql_expr
        else:
            new_query.where = LogicalBinaryExpression(
                type_=LogicalBinaryType.AND,
                left=new_query.where,
                right=sql_expr
            )
        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        pure_expr = self.__filter_expr.to_pure_expression(config)
        return f"{self.__base_frame.to_pure(config)}{config.separator(1)}->filter(c|{pure_expr})"

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import collections.abc
from pylegend._typing import (
    PyLegendCallable,
    PyLegendSequence,
    PyLegendTuple,
    PyLegendUnion,
    PyLegendList,
    PyLegendMapping,
)
from pylegend.core.language.pandas_api.pandas_api_aggregate_specification import (
    PyLegendAggFunc,
    PyLegendAggInput,
    PyLegendAggList,
)
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.helpers import escape_column_name, generate_pure_lambda
from pylegend.core.language.shared.literal_expressions import convert_literal_to_literal_expression
from pylegend.core.language.shared.primitive_collection import PyLegendPrimitiveCollection, create_primitive_collection
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.primitives.date import PyLegendDate
from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
from pylegend.core.language.shared.primitives.float import PyLegendFloat
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.number import PyLegendNumber
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitive, PyLegendPrimitiveOrPythonPrimitive
from pylegend.core.language.shared.primitives.strictdate import PyLegendStrictDate
from pylegend.core.language.shared.primitives.string import PyLegendString
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SelectItem,
    SingleColumn,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.tds.tds_column import PrimitiveTdsColumn, TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig, FrameToSqlConfig


class AggregateFunction(PandasApiAppliedFunction):
    __base_frame: PyLegendUnion[PandasApiBaseTdsFrame, PandasApiGroupbyTdsFrame]
    __func: PyLegendAggInput
    __axis: PyLegendUnion[int, str]
    __args: PyLegendSequence[PyLegendPrimitiveOrPythonPrimitive]
    __kwargs: PyLegendMapping[str, PyLegendPrimitiveOrPythonPrimitive]

    @classmethod
    def name(cls) -> str:
        return "aggregate"  # pragma: no cover

    def __init__(
        self,
        base_frame: PyLegendUnion[PandasApiBaseTdsFrame, PandasApiGroupbyTdsFrame],
        func: PyLegendAggInput,
        axis: PyLegendUnion[int, str],
        *args: PyLegendPrimitiveOrPythonPrimitive,
        **kwargs: PyLegendPrimitiveOrPythonPrimitive,
    ) -> None:
        self.__base_frame = base_frame
        self.__func = func
        self.__axis = axis
        self.__args = args
        self.__kwargs = kwargs

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        db_extension = config.sql_to_string_generator().get_db_extension()

        base_query: QuerySpecification = self.base_frame().to_sql_query_object(config)

        should_create_sub_query = (
            len(base_query.groupBy) > 0
            or base_query.select.distinct
            or base_query.offset is not None
            or base_query.limit is not None
        )

        new_query: QuerySpecification
        if should_create_sub_query:
            new_query = create_sub_query(base_query, config, "root")
        else:
            new_query = copy_query(base_query)

        new_select_items: PyLegendList[SelectItem] = []

        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            columns_to_retain: PyLegendList[str] = [
                db_extension.quote_identifier(x) for x in self.__base_frame.grouping_column_name_list()
            ]
            new_cols_with_index: PyLegendList[PyLegendTuple[int, "SelectItem"]] = []
            for col in new_query.select.selectItems:
                if not isinstance(col, SingleColumn):
                    raise ValueError(
                        "Group By operation not supported for queries " "with columns other than SingleColumn"
                    )  # pragma: no cover
                if col.alias is None:
                    raise ValueError(
                        "Group By operation not supported for queries " "with SingleColumns with missing alias"
                    )  # pragma: no cover
                if col.alias in columns_to_retain:
                    new_cols_with_index.append((columns_to_retain.index(col.alias), col))

            new_select_items = [y[1] for y in sorted(new_cols_with_index, key=lambda x: x[0])]

        for agg in self.__aggregates_list:
            agg_sql_expr = agg[2].to_sql_expression({"r": new_query}, config)

            new_select_items.append(SingleColumn(alias=db_extension.quote_identifier(agg[0]), expression=agg_sql_expr))

        new_query.select.selectItems = new_select_items

        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            tds_row = PandasApiTdsRow.from_tds_frame("r", self.base_frame())
            new_query.groupBy = [
                (lambda x: x[c])(tds_row).to_sql_expression({"r": new_query}, config)
                for c in self.__base_frame.grouping_column_name_list()
            ]

        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        agg_strings = []
        for agg in self.__aggregates_list:
            map_expr_string = (
                agg[1].to_pure_expression(config)
                if isinstance(agg[1], PyLegendPrimitive)
                else convert_literal_to_literal_expression(agg[1]).to_pure_expression(config)
            )
            agg_expr_string = agg[2].to_pure_expression(config).replace(map_expr_string, "$c")
            agg_strings.append(
                f"{escape_column_name(agg[0])}:{generate_pure_lambda('r', map_expr_string)}:"
                f"{generate_pure_lambda('c', agg_expr_string)}"
            )

        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            group_strings = []
            for col_name in self.__base_frame.grouping_column_name_list():
                group_strings.append(escape_column_name(col_name))

            pure_expression = (
                f"{self.base_frame().to_pure(config)}{config.separator(1)}" + f"->groupBy({config.separator(2)}"
                f"~[{', '.join(group_strings)}],{config.separator(2, True)}"
                f"~[{', '.join(agg_strings)}]{config.separator(1)}"
                f")"
            )

            return pure_expression
        else:
            return (
                f"{self.__base_frame.to_pure(config)}{config.separator(1)}"
                f"->aggregate({config.separator(2)}"
                f"~[{', '.join(agg_strings)}]{config.separator(1)}"
                f")"
            )

    def base_frame(self) -> PandasApiBaseTdsFrame:
        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            return self.__base_frame.base_frame()
        else:
            return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        new_columns = []

        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            base_cols_map = {c.get_name(): c for c in self.base_frame().columns()}
            for group_col_name in self.__base_frame.grouping_column_name_list():
                if group_col_name in base_cols_map:
                    new_columns.append(base_cols_map[group_col_name].copy())

        for alias, _, agg_expr in self.__aggregates_list:
            new_columns.append(self.__infer_column_from_expression(alias, agg_expr))

        return new_columns

    def __infer_column_from_expression(self, name: str, expr: PyLegendPrimitive) -> TdsColumn:
        if isinstance(expr, PyLegendInteger):
            return PrimitiveTdsColumn.integer_column(name)
        elif isinstance(expr, PyLegendFloat):
            return PrimitiveTdsColumn.float_column(name)
        elif isinstance(expr, PyLegendNumber):
            return PrimitiveTdsColumn.number_column(name)
        elif isinstance(expr, PyLegendString):
            return PrimitiveTdsColumn.string_column(name)
        elif isinstance(expr, PyLegendBoolean):
            return PrimitiveTdsColumn.boolean_column(name)  # pragma: no cover
        elif isinstance(expr, PyLegendDate):
            return PrimitiveTdsColumn.date_column(name)
        elif isinstance(expr, PyLegendDateTime):
            return PrimitiveTdsColumn.datetime_column(name)
        elif isinstance(expr, PyLegendStrictDate):
            return PrimitiveTdsColumn.strictdate_column(name)
        else:
            raise TypeError(f"Could not infer TdsColumn type for aggregation result type: {type(expr)}")  # pragma: no cover

    def validate(self) -> bool:
        if self.__axis not in [0, "index"]:
            raise NotImplementedError(
                f"The 'axis' parameter of the aggregate function must be 0 or 'index', but got: {self.__axis}"
            )

        if len(self.__args) > 0 or len(self.__kwargs) > 0:
            raise NotImplementedError(
                "AggregateFunction currently does not support additional positional "
                "or keyword arguments. Please remove extra *args/**kwargs."
            )

        self.__aggregates_list: PyLegendList[PyLegendTuple[str, PyLegendPrimitiveOrPythonPrimitive, PyLegendPrimitive]] = []

        normalized_func: dict[str, PyLegendUnion[PyLegendAggFunc, PyLegendAggList]] = (
            self.__normalize_input_func_to_standard_dict(self.__func)
        )

        tds_row = PandasApiTdsRow.from_tds_frame("r", self.base_frame())

        for column_name, agg_input in normalized_func.items():
            mapper_function: PyLegendCallable[[PandasApiTdsRow], PyLegendPrimitiveOrPythonPrimitive] = eval(
                f'lambda r: r["{column_name}"]'
            )
            map_result: PyLegendPrimitiveOrPythonPrimitive = mapper_function(tds_row)
            collection: PyLegendPrimitiveCollection = create_primitive_collection(map_result)

            if isinstance(agg_input, list):
                lambda_counter = 0
                for func in agg_input:
                    is_anonymous_lambda = False
                    if not isinstance(func, str):
                        if getattr(func, "__name__", "<lambda>") == "<lambda>":
                            is_anonymous_lambda = True

                    if is_anonymous_lambda:
                        lambda_counter += 1

                    normalized_agg_func = self.__normalize_agg_func_to_lambda_function(func)
                    agg_result = normalized_agg_func(collection)

                    alias = self._generate_column_alias(column_name, func, lambda_counter)
                    self.__aggregates_list.append((alias, map_result, agg_result))

            else:
                normalized_agg_func = self.__normalize_agg_func_to_lambda_function(agg_input)
                agg_result = normalized_agg_func(collection)

                self.__aggregates_list.append((column_name, map_result, agg_result))

        return True

    def __normalize_input_func_to_standard_dict(
        self, func_input: PyLegendAggInput
    ) -> dict[str, PyLegendUnion[PyLegendAggFunc, PyLegendAggList]]:

        validation_columns: PyLegendList[str]
        default_broadcast_columns: PyLegendList[str]
        group_cols: set[str] = set()

        all_cols = [col.get_name() for col in self.base_frame().columns()]

        if isinstance(self.__base_frame, PandasApiGroupbyTdsFrame):
            group_cols = set(self.__base_frame.grouping_column_name_list())

            selected_cols = self.__base_frame.selected_columns()

            if selected_cols is not None:
                validation_columns = selected_cols
                default_broadcast_columns = selected_cols
            else:
                validation_columns = all_cols
                default_broadcast_columns = [c for c in all_cols if c not in group_cols]
        else:
            validation_columns = all_cols
            default_broadcast_columns = all_cols

        if isinstance(func_input, collections.abc.Mapping):
            normalized: dict[str, PyLegendUnion[PyLegendAggFunc, PyLegendAggList]] = {}

            for key, value in func_input.items():
                if not isinstance(key, str):
                    raise TypeError(
                        f"Invalid `func` argument for the aggregate function.\n"
                        f"When a dictionary is provided, all keys must be strings.\n"
                        f"But got key: {key!r} (type: {type(key).__name__})\n"
                    )

                if key not in validation_columns:
                    raise ValueError(
                        f"Invalid `func` argument for the aggregate function.\n"
                        f"When a dictionary is provided, all keys must be column names.\n"
                        f"Available columns are: {sorted(validation_columns)}\n"
                        f"But got key: {key!r} (type: {type(key).__name__})\n"
                    )

                if isinstance(value, collections.abc.Sequence) and not isinstance(value, str):
                    for i, f in enumerate(value):
                        if not (callable(f) or isinstance(f, str) or isinstance(f, np.ufunc)):
                            raise TypeError(
                                f"Invalid `func` argument for the aggregate function.\n"
                                f"When a list is provided for a column, all elements must be callable, str, or np.ufunc.\n"
                                f"But got element at index {i}: {f!r} (type: {type(f).__name__})\n"
                            )
                    normalized[key] = value

                else:
                    if not (callable(value) or isinstance(value, str) or isinstance(value, np.ufunc)):
                        raise TypeError(
                            f"Invalid `func` argument for the aggregate function.\n"
                            f"When a dictionary is provided, the value must be a callable, str, or np.ufunc "
                            f"(or a list containing these).\n"
                            f"But got value for key '{key}': {value} (type: {type(value).__name__})\n"
                        )

                    if key in group_cols:
                        normalized[key] = [value]
                    else:
                        normalized[key] = value

            return normalized

        elif isinstance(func_input, collections.abc.Sequence) and not isinstance(func_input, str):
            for i, f in enumerate(func_input):
                if not (callable(f) or isinstance(f, str) or isinstance(f, np.ufunc)):
                    raise TypeError(
                        f"Invalid `func` argument for the aggregate function.\n"
                        f"When a list is provided as the main argument, all elements must be callable, str, or np.ufunc.\n"
                        f"But got element at index {i}: {f!r} (type: {type(f).__name__})\n"
                    )

            return {col: func_input for col in default_broadcast_columns}

        elif callable(func_input) or isinstance(func_input, str) or isinstance(func_input, np.ufunc):
            return {col: func_input for col in default_broadcast_columns}

        else:
            raise TypeError(
                "Invalid `func` argument for aggregate function. "
                "Expected a callable, str, np.ufunc, a list containing exactly one of these, "
                "or a mapping[str -> callable/str/ufunc/a list containing exactly one of these]. "
                f"But got: {func_input!r} (type: {type(func_input).__name__})"
            )

    def __normalize_agg_func_to_lambda_function(
        self, func: PyLegendAggFunc
    ) -> PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]:

        PYTHON_FUNCTION_TO_LEGEND_FUNCTION_MAPPING: PyLegendMapping[str, PyLegendList[str]] = {
            "average": ["mean", "average", "nanmean"],
            "sum": ["sum", "nansum"],
            "min": ["min", "amin", "minimum", "nanmin"],
            "max": ["max", "amax", "maximum", "nanmax"],
            "std_dev_sample": ["std", "std_dev", "nanstd"],
            "variance_sample": ["var", "variance", "nanvar"],
            "count": ["count", "size", "len", "length"],
        }

        FLATTENED_FUNCTION_MAPPING: dict[str, str] = {}
        for target_method, source_list in PYTHON_FUNCTION_TO_LEGEND_FUNCTION_MAPPING.items():
            for alias in source_list:
                FLATTENED_FUNCTION_MAPPING[alias] = target_method

        lambda_source: str
        final_lambda: PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]

        if isinstance(func, str):
            func_lower = func.lower()
            if func_lower in FLATTENED_FUNCTION_MAPPING:
                internal_method_name = FLATTENED_FUNCTION_MAPPING[func_lower]
            else:
                raise NotImplementedError(
                    f"Invalid `func` argument for the aggregate function.\n"
                    f"The string {func!r} does not correspond to any supported aggregation.\n"
                    f"Available string functions are: {sorted(FLATTENED_FUNCTION_MAPPING.keys())}"
                )  # pragma: no cover
            lambda_source = self._generate_lambda_source(internal_method_name)
            final_lambda = eval(lambda_source)
            return final_lambda

        elif isinstance(func, np.ufunc):
            func_name = func.__name__
            if func_name in FLATTENED_FUNCTION_MAPPING:
                internal_method_name = FLATTENED_FUNCTION_MAPPING[func_name]
            else:
                raise NotImplementedError(
                    f"Invalid `func` argument for the aggregate function.\n"
                    f"The NumPy function {func_name!r} is not supported.\n"
                    f"Supported aggregate functions are: {sorted(FLATTENED_FUNCTION_MAPPING.keys())}"
                )  # pragma: no cover
            lambda_source = self._generate_lambda_source(internal_method_name)
            final_lambda = eval(lambda_source)
            return final_lambda

        else:
            func_name = getattr(func, "__name__", "").lower()
            if func_name in FLATTENED_FUNCTION_MAPPING and func_name != "<lambda>":
                internal_method_name = FLATTENED_FUNCTION_MAPPING[func_name]
                lambda_source = self._generate_lambda_source(internal_method_name)
                final_lambda = eval(lambda_source)
                return final_lambda
            else:

                def validation_wrapper(x: PyLegendPrimitiveCollection) -> PyLegendPrimitive:
                    result = func(x)
                    if not isinstance(result, PyLegendPrimitive):
                        raise TypeError(
                            f"Custom aggregation function must return a PyLegendPrimitive (Expression).\n"
                            f"But got type: {type(result).__name__}\n"
                            f"Value: {result!r}"
                        )  # pragma: no cover
                    return result

                return validation_wrapper

    def _generate_lambda_source(self, internal_method_name: str) -> str:
        return f"lambda x: x.{internal_method_name}()"

    def _generate_column_alias(self, col_name: str, func: PyLegendAggFunc, lambda_counter: int) -> str:
        if isinstance(func, str):
            return f"{func}({col_name})"

        func_name = getattr(func, "__name__", "<lambda>")

        if func_name != "<lambda>":
            return f"{func_name}({col_name})"
        else:
            return f"lambda_{lambda_counter}({col_name})"



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
)
from pylegend.core.sql.metamodel import (
    Expression,
    Window,
)

__all__: PyLegendSequence[str] = [
    "StringLengthExpression",
    "StringLikeExpression",
    "StringUpperExpression",
    "StringLowerExpression",
    "TrimType",
    "StringTrimExpression",
    "StringPosExpression",
    "StringConcatExpression",
    "AbsoluteExpression",
    "PowerExpression",
    "CeilExpression",
    "FloorExpression",
    "SqrtExpression",
    "CbrtExpression",
    "ExpExpression",
    "LogExpression",
    "RemainderExpression",
    "RoundExpression",
    "SineExpression",
    "ArcSineExpression",
    "CosineExpression",
    "ArcCosineExpression",
    "TanExpression",
    "ArcTanExpression",
    "ArcTan2Expression",
    "CotExpression",
    "CountExpression",
    "DistinctCountExpression",
    "AverageExpression",
    "MaxExpression",
    "MinExpression",
    "SumExpression",
    "StdDevSampleExpression",
    "StdDevPopulationExpression",
    "VarianceSampleExpression",
    "VariancePopulationExpression",
    "JoinStringsExpression",
    "FirstDayOfYearExpression",
    "FirstDayOfQuarterExpression",
    "FirstDayOfMonthExpression",
    "FirstDayOfWeekExpression",
    "FirstHourOfDayExpression",
    "FirstMinuteOfHourExpression",
    "FirstSecondOfMinuteExpression",
    "FirstMillisecondOfSecondExpression",
    "YearExpression",
    "QuarterExpression",
    "MonthExpression",
    "WeekOfYearExpression",
    "DayOfYearExpression",
    "DayOfMonthExpression",
    "DayOfWeekExpression",
    "HourExpression",
    "MinuteExpression",
    "SecondExpression",
    "EpochExpression",
    "WindowExpression",
    "ConstantExpression",
    "StringSubStringExpression"
]


class StringLengthExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression"
    ) -> None:
        super().__init__(_type="stringLengthExpression")
        self.value = value


class StringLikeExpression(Expression):
    value: "Expression"
    other: "Expression"

    def __init__(
        self,
        value: "Expression",
        other: "Expression"
    ) -> None:
        super().__init__(_type="stringLikeExpression")
        self.value = value
        self.other = other


class StringUpperExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression"
    ) -> None:
        super().__init__(_type="stringUpperExpression")
        self.value = value


class StringLowerExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression"
    ) -> None:
        super().__init__(_type="stringLowerExpression")
        self.value = value


class TrimType(Enum):
    Left = 1,
    Right = 2,
    Both = 3


class StringTrimExpression(Expression):
    value: "Expression"
    trim_type: TrimType

    def __init__(
        self,
        value: "Expression",
        trim_type: TrimType
    ) -> None:
        super().__init__(_type="stringTrimExpression")
        self.value = value
        self.trim_type = trim_type


class StringPosExpression(Expression):
    value: "Expression"
    other: "Expression"

    def __init__(
        self,
        value: "Expression",
        other: "Expression"
    ) -> None:
        super().__init__(_type="stringPosExpression")
        self.value = value
        self.other = other


class StringConcatExpression(Expression):
    first: "Expression"
    second: "Expression"

    def __init__(
        self,
        first: "Expression",
        second: "Expression"
    ) -> None:
        super().__init__(_type="stringConcatExpression")
        self.first = first
        self.second = second


class AbsoluteExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="absoluteExpression")
        self.value = value


class PowerExpression(Expression):
    first: "Expression"
    second: "Expression"

    def __init__(
        self,
        first: "Expression",
        second: "Expression"
    ) -> None:
        super().__init__(_type="powerExpression")
        self.first = first
        self.second = second


class CeilExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="ceilExpression")
        self.value = value


class FloorExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="floorExpression")
        self.value = value


class SqrtExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="sqrtExpression")
        self.value = value


class CbrtExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="cbrtExpression")
        self.value = value


class ExpExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="expExpression")
        self.value = value


class LogExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="logExpression")
        self.value = value


class RemainderExpression(Expression):
    first: "Expression"
    second: "Expression"

    def __init__(
        self,
        first: "Expression",
        second: "Expression"
    ) -> None:
        super().__init__(_type="remainderExpression")
        self.first = first
        self.second = second


class RoundExpression(Expression):
    first: "Expression"
    second: "PyLegendOptional[Expression]"

    def __init__(
        self,
        first: "Expression",
        second: "PyLegendOptional[Expression]"
    ) -> None:
        super().__init__(_type="roundExpression")
        self.first = first
        self.second = second


class SineExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="sineExpression")
        self.value = value


class ArcSineExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="arcSineExpression")
        self.value = value


class CosineExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="cosineExpression")
        self.value = value


class ArcCosineExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="arcCosineExpression")
        self.value = value


class TanExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="tanExpression")
        self.value = value


class ArcTanExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="arcTanExpression")
        self.value = value


class ArcTan2Expression(Expression):
    first: "Expression"
    second: "Expression"

    def __init__(
        self,
        first: "Expression",
        second: "Expression"
    ) -> None:
        super().__init__(_type="arcTan2Expression")
        self.first = first
        self.second = second


class CotExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="cotExpression")
        self.value = value


class CountExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="countExpression")
        self.value = value


class DistinctCountExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="distinctCountExpression")
        self.value = value


class AverageExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="averageExpression")
        self.value = value


class MaxExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="maxExpression")
        self.value = value


class MinExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="minExpression")
        self.value = value


class SumExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="sumExpression")
        self.value = value


class StdDevSampleExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="stdDevSampleExpression")
        self.value = value


class StdDevPopulationExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="stdDevPopulationExpression")
        self.value = value


class VarianceSampleExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="varianceSampleExpression")
        self.value = value


class VariancePopulationExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="variancePopulationExpression")
        self.value = value


class JoinStringsExpression(Expression):
    value: "Expression"
    other: "Expression"

    def __init__(
        self,
        value: "Expression",
        other: "Expression",
    ) -> None:
        super().__init__(_type="joinStringsExpression")
        self.value = value
        self.other = other


class FirstDayOfYearExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="firstDayOfYearExpression")
        self.value = value


class FirstDayOfQuarterExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="firstDayOfQuarterExpression")
        self.value = value


class FirstDayOfMonthExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="firstDayOfMonthExpression")
        self.value = value


class FirstDayOfWeekExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="firstDayOfWeekExpression")
        self.value = value


class FirstHourOfDayExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="firstHourOfDayExpression")
        self.value = value


class FirstMinuteOfHourExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="firstMinuteOfHourExpression")
        self.value = value


class FirstSecondOfMinuteExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="firstSecondOfMinuteExpression")
        self.value = value


class FirstMillisecondOfSecondExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="firstMillisecondOfSecondExpression")
        self.value = value


class YearExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="yearExpression")
        self.value = value


class QuarterExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="quarterExpression")
        self.value = value


class MonthExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="monthExpression")
        self.value = value


class WeekOfYearExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="weekOfYearExpression")
        self.value = value


class DayOfYearExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="dayOfYearExpression")
        self.value = value


class DayOfMonthExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="dayOfMonthExpression")
        self.value = value


class DayOfWeekExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="dayOfWeekExpression")
        self.value = value


class HourExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="hourExpression")
        self.value = value


class MinuteExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="minuteExpression")
        self.value = value


class SecondExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="secondExpression")
        self.value = value


class EpochExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression",
    ) -> None:
        super().__init__(_type="epochExpression")
        self.value = value


class WindowExpression(Expression):
    nested: "Expression"
    window: "Window"

    def __init__(
        self,
        nested: "Expression",
        window: "Window",
    ) -> None:
        super().__init__(_type="windowExpression")
        self.nested = nested
        self.window = window


class ConstantExpression(Expression):
    name: str

    def __init__(
        self,
        name: str
    ) -> None:
        super().__init__(_type="constantExpression")
        self.name = name


class StringSubStringExpression(Expression):
    value: "Expression"
    start: "Expression"
    end: PyLegendOptional["Expression"]

    def __init__(
            self,
            value: "Expression",
            start: "Expression",
            end: PyLegendOptional["Expression"] = None
    ) -> None:
        super().__init__(_type="stringSubStringExpression")
        self.value = value
        self.start = start
        self.end = end



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.language.shared.primitives import (
    PyLegendPrimitive,
    PyLegendPrimitiveOrPythonPrimitive,
    PyLegendBoolean,
    PyLegendString,
    PyLegendNumber,
    PyLegendInteger,
    PyLegendFloat,
    PyLegendDate,
    PyLegendDateTime,
    PyLegendStrictDate,
)
from pylegend.core.language.shared.expression import (
    PyLegendExpressionBooleanReturn,
    PyLegendExpressionStringReturn,
    PyLegendExpressionNumberReturn,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionFloatReturn,
    PyLegendExpressionDateReturn,
    PyLegendExpressionDateTimeReturn,
    PyLegendExpressionStrictDateReturn,
)
from pylegend.core.language.shared.literal_expressions import (
    PyLegendBooleanLiteralExpression,
    PyLegendStringLiteralExpression,
    PyLegendIntegerLiteralExpression,
    PyLegendFloatLiteralExpression,
    PyLegendDateTimeLiteralExpression,
    PyLegendStrictDateLiteralExpression,
    convert_literal_to_literal_expression,
)
from pylegend.core.language.shared.column_expressions import (
    PyLegendColumnExpression,
    PyLegendBooleanColumnExpression,
    PyLegendStringColumnExpression,
    PyLegendNumberColumnExpression,
    PyLegendIntegerColumnExpression,
    PyLegendFloatColumnExpression,
    PyLegendDateColumnExpression,
    PyLegendDateTimeColumnExpression,
    PyLegendStrictDateColumnExpression,
)
from pylegend.core.language.legacy_api.legacy_api_tds_row import LegacyApiTdsRow
from pylegend.core.language.legacy_api.aggregate_specification import LegacyApiAggregateSpecification, agg
from pylegend.core.language.shared.primitive_collection import (
    PyLegendPrimitiveCollection,
    PyLegendIntegerCollection,
    PyLegendFloatCollection,
    PyLegendNumberCollection,
    PyLegendStringCollection,
    PyLegendBooleanCollection,
    PyLegendDateCollection,
    PyLegendDateTimeCollection,
    PyLegendStrictDateCollection,
    create_primitive_collection,
)
from pylegend.core.language.shared.functions import (
    today,
    now,
    current_user,
)

__all__: PyLegendSequence[str] = [
    "PyLegendPrimitive",
    "PyLegendPrimitiveOrPythonPrimitive",
    "PyLegendBoolean",
    "PyLegendString",
    "PyLegendNumber",
    "PyLegendInteger",
    "PyLegendFloat",
    "PyLegendDate",
    "PyLegendDateTime",
    "PyLegendStrictDate",

    "PyLegendExpressionBooleanReturn",
    "PyLegendExpressionStringReturn",
    "PyLegendExpressionNumberReturn",
    "PyLegendExpressionIntegerReturn",
    "PyLegendExpressionFloatReturn",
    "PyLegendExpressionDateReturn",
    "PyLegendExpressionDateTimeReturn",
    "PyLegendExpressionStrictDateReturn",

    "PyLegendColumnExpression",
    "PyLegendBooleanColumnExpression",
    "PyLegendStringColumnExpression",
    "PyLegendNumberColumnExpression",
    "PyLegendIntegerColumnExpression",
    "PyLegendFloatColumnExpression",
    "PyLegendDateColumnExpression",
    "PyLegendDateTimeColumnExpression",
    "PyLegendStrictDateColumnExpression",

    "PyLegendBooleanLiteralExpression",
    "PyLegendStringLiteralExpression",
    "PyLegendIntegerLiteralExpression",
    "PyLegendFloatLiteralExpression",
    "PyLegendDateTimeLiteralExpression",
    "PyLegendStrictDateLiteralExpression",
    "convert_literal_to_literal_expression",

    "LegacyApiTdsRow",
    "LegacyApiAggregateSpecification",
    "agg",

    "PyLegendPrimitiveCollection",
    "PyLegendIntegerCollection",
    "PyLegendFloatCollection",
    "PyLegendNumberCollection",
    "PyLegendStringCollection",
    "PyLegendBooleanCollection",
    "PyLegendDateCollection",
    "PyLegendDateTimeCollection",
    "PyLegendStrictDateCollection",
    "create_primitive_collection",

    "today",
    "now",
    "current_user",
]



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import date
from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendUnion,
    PyLegendTuple,
    PyLegendOptional,
)
from pylegend.core.sql.metamodel import LongLiteral, QuerySpecification
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.sql_query_helpers import create_sub_query, copy_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig, FrameToSqlConfig


class TruncateFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __before: int
    __after: PyLegendUnion[int, None]
    __axis: PyLegendUnion[str, int]
    __copy: bool

    @classmethod
    def name(cls) -> str:
        return "truncate"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            before: PyLegendUnion[date, str, int, None],
            after: PyLegendUnion[date, str, int, None],
            axis: PyLegendUnion[str, int],
            copy: bool,
    ) -> None:
        self.__base_frame = base_frame
        self.__before_input = before
        self.__after_input = after
        self.__axis = axis
        self.__copy = copy

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query: QuerySpecification = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (base_query.offset is not None) or (base_query.limit is not None)
        new_query = create_sub_query(base_query, config, "root") if should_create_sub_query else copy_query(base_query)
        new_query.offset = LongLiteral(self.__before)

        if self.__after is not None:
            new_query.limit = LongLiteral(self.__after - self.__before + 1)
        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        if self.__after is None:
            return f"{self.__base_frame.to_pure(config)}{config.separator(1)}" f"->drop({self.__before})"

        start_row = self.__before
        end_row = self.__after + 1
        return f"{self.__base_frame.to_pure(config)}{config.separator(1)}" f"->slice({start_row}, {end_row})"

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        if self.__axis not in [0, "index"]:
            raise NotImplementedError(
                f"The 'axis' parameter of the truncate function must be 0 or 'index', but got: {self.__axis}"
            )

        if self.__copy not in [True]:
            raise NotImplementedError(f"The 'copy' parameter of the truncate function must be True, but got: {self.__copy}")

        self.__before, self.__after = self.__normalize_before_and_after(self.__before_input, self.__after_input)
        return True

    @staticmethod
    def __normalize_before_and_after(
            before_input: PyLegendUnion[date, str, int, None],
            after_input: PyLegendUnion[date, str, int, None]
    ) -> PyLegendTuple[int, PyLegendOptional[int]]:

        if isinstance(before_input, (date, str)):
            raise NotImplementedError(
                f"The 'before' parameter of the truncate function must be of type integer or None, "
                f"but got: before={before_input} (type: {type(before_input).__name__})")

        if isinstance(after_input, (date, str)):
            raise NotImplementedError(
                f"The 'after' parameter of the truncate function must be of type integer or None, "
                f"but got: after={after_input} (type: {type(after_input).__name__})")

        def __raise_error_if_before_gt_after(before_input: int, after_input: int) -> None:
            if before_input > after_input:
                raise ValueError(
                    f"The 'before' parameter of the truncate function must be less than or equal to the 'after' parameter, "
                    f"but got: before={before_input}, after={after_input}")

        if before_input is None:
            if after_input is None:
                return 0, None

            if isinstance(after_input, int) and after_input >= 0:
                return 0, after_input

            if isinstance(after_input, int) and after_input < 0:
                return 0, -1

        if isinstance(before_input, int) and before_input >= 0:
            if after_input is None:
                return before_input, None

            if isinstance(after_input, int) and after_input >= 0:
                __raise_error_if_before_gt_after(before_input, after_input)
                return before_input, after_input

            if isinstance(after_input, int) and after_input < 0:
                __raise_error_if_before_gt_after(before_input, after_input)

        if isinstance(before_input, int) and before_input < 0:
            if after_input is None:
                return 0, None

            if isinstance(after_input, int) and after_input >= 0:
                __raise_error_if_before_gt_after(before_input, after_input)
                return 0, after_input

            if isinstance(after_input, int) and after_input < 0:
                __raise_error_if_before_gt_after(before_input, after_input)
                return 0, -1

        return 0, 0  # pragma: no cover



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABCMeta, abstractmethod
from datetime import date, datetime
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
    TYPE_CHECKING,
)
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.language.shared.expression import PyLegendExpression
from pylegend.core.language.shared.literal_expressions import convert_literal_to_literal_expression
from pylegend.core.language.shared.operations.primitive_operation_expressions import (
    PyLegendPrimitiveEqualsExpression,
    PyLegendPrimitiveNotEqualsExpression,
    PyLegendIsEmptyExpression,
    PyLegendIsNotEmptyExpression,
    PyLegendPrimitiveToStringExpression
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
if TYPE_CHECKING:
    from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
    from pylegend.core.language.shared.primitives.string import PyLegendString

__all__: PyLegendSequence[str] = [
    "PyLegendPrimitive",
    "PyLegendPrimitiveOrPythonPrimitive",
]


class PyLegendPrimitive(metaclass=ABCMeta):

    @abstractmethod
    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        pass

    @abstractmethod
    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        pass

    def __eq__(  # type: ignore
            self,
            other: "PyLegendUnion[int, float, bool, str, date, datetime, PyLegendPrimitive]"
    ) -> "PyLegendBoolean":
        PyLegendPrimitive.__validate_param_to_be_primitive(other, "Equals (==) parameter")

        if isinstance(other, (int, float, bool, str, date, datetime)):
            other_op = convert_literal_to_literal_expression(other)
        else:
            other_op = other.value()

        from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
        return PyLegendBoolean(PyLegendPrimitiveEqualsExpression(self.value(), other_op))

    def __ne__(  # type: ignore
            self,
            other: "PyLegendUnion[int, float, bool, str, date, datetime, PyLegendPrimitive]"
    ) -> "PyLegendBoolean":
        PyLegendPrimitive.__validate_param_to_be_primitive(other, "Not Equals (!=) parameter")

        if isinstance(other, (int, float, bool, str, date, datetime)):
            other_op = convert_literal_to_literal_expression(other)
        else:
            other_op = other.value()

        from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
        return PyLegendBoolean(PyLegendPrimitiveNotEqualsExpression(self.value(), other_op))

    def is_empty(self) -> "PyLegendBoolean":
        from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
        return PyLegendBoolean(PyLegendIsEmptyExpression(self.value()))

    def is_null(self) -> "PyLegendBoolean":
        return self.is_empty()

    def is_not_empty(self) -> "PyLegendBoolean":
        from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
        return PyLegendBoolean(PyLegendIsNotEmptyExpression(self.value()))

    def is_not_null(self) -> "PyLegendBoolean":
        return self.is_not_empty()

    def to_string(self) -> "PyLegendString":
        from pylegend.core.language.shared.primitives.string import PyLegendString
        return PyLegendString(PyLegendPrimitiveToStringExpression(self.value()))

    @staticmethod
    def __validate_param_to_be_primitive(
            param: "PyLegendUnion[int, float, bool, str, date, datetime, PyLegendPrimitive]",
            desc: str
    ) -> None:
        if not isinstance(param, (int, float, bool, str, date, datetime, PyLegendPrimitive)):
            raise TypeError(
                desc + " should be a int/float/bool/str/datetime.date/datetime.datetime or a primitive expression."
                       " Got value " + str(param) + " of type: " + str(type(param)))

    @abstractmethod
    def value(self) -> PyLegendExpression:
        pass


PyLegendPrimitiveOrPythonPrimitive = PyLegendUnion[int, float, str, bool, date, datetime, PyLegendPrimitive]
